import numpy as np 
from statsmodels.distributions.empirical_distribution import ECDF

def Sharpe_rat(risk_returns,safe_asset,Rebalancement_frequency):

        """
        Compute the Sharpe Ratio

        Parameters
        ----------
        risk_returns : np.ndarray
        safe_asset : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, Sharpe Ratio

        """

        Excess_returns = np.mean(risk_returns - safe_asset/Rebalancement_frequency) * Rebalancement_frequency
        strat_std = np.std(risk_returns) * Rebalancement_frequency**.5
        Sharpe_ratio = Excess_returns/strat_std
        return Sharpe_ratio
                          
def Sortino_rat(risk_returns,safe_asset,Rebalancement_frequency):

        """
        Compute the Sortino Ratio

        Parameters
        ----------
        risk_returns : np.ndarray
        safe_asset : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, Sortino Ratio

        """

        Sortino_std = np.std(risk_returns[np.where(risk_returns > 0 )]) * Rebalancement_frequency**.5
        Excess_returns = np.mean(risk_returns- safe_asset / Rebalancement_frequency) * Rebalancement_frequency
        Sortino_ratio = Excess_returns/Sortino_std
        return Sortino_ratio
                          
def Information_rat(risk_returns, benchmark_returns, Rebalancement_frequency):

        """
        Compute the breadth

        Parameters
        ----------
        risk_returns : np.ndarray
        benchmark_returns : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, Information Ratio

        """

        differentials = risk_returns - benchmark_returns 
        volatility = np.std(differentials) * Rebalancement_frequency**.5
        IR = np.mean(differentials) * Rebalancement_frequency / volatility
        return IR
    
def Modigliani_rat(risk_returns, benchmark_returns, safe_asset, Rebalancement_frequency,Sharpe_ratio):

        """
        Compute the Modigliani Ratio 

        Parameters
        ----------
        risk_returns : np.ndarray
        benchmark_returns : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, Modigliani Ratio 

        """

        benchmark_volatility = np.std(benchmark_returns) * Rebalancement_frequency**.5
        m2_ratio = Sharpe_ratio * benchmark_volatility + safe_asset[-1]
        return m2_ratio


def Omega_rat(risk_returns):

        """
        Compute the Omega ratio 

        Parameters
        ----------
        risk_returns : np.ndarray

        Returns
        ----------

        float64, Omega ratio 

        """

        ecdf = ECDF(risk_returns)
        Omega_ratio = (1-ecdf(0))/ecdf(0)
        return Omega_ratio


import os

import pandas as pd
from pandas import DataFrame

def load_file(file_base: str, filename: str) -> DataFrame:
    """
    Load data from a csv.gz file.
    Parameters
    ----------
    file_base : str
        The director to use as the root.
    filename : str
        Name of csv.gz to load
    Returns
    -------
    DataFrame
        Dataframe containing the loaded data.
    """
    curr_dir = os.path.split(os.path.abspath(file_base))[0]
    data = pd.read_csv(os.path.join(curr_dir, filename))
    if "Date" in data:
        data.Date = pd.to_datetime(data.Date)
        data = data.set_index("Date")
    for col in data:
        data[col] = pd.to_numeric(data[col], errors="coerce")
    return data


import numpy as np 

def stdeviation(risk_returns,Rebalancement_frequency):

        """
        Compute the standard deviation of the investment  

        Parameters
        ----------
        risk_returns : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, standard deviation 

        """

        stdev = np.std(risk_returns) * Rebalancement_frequency**.5
        return stdev


import numpy as np
import numba 
from numba import njit

class TIPP :
    
    def __init__(self,risk_returns,safe_asset,Lock_in,\
                 Min_risk_part,Capital_reinjection_rate,Initial_funds\
                ,floor_percent,multiplier,Rebalancement_frequency = 252):

            self.risk_returns = risk_returns
            self.Rebalancement_frequency = Rebalancement_frequency
            self.safe_asset = safe_asset
            self.Lock_in = Lock_in
            self.Min_risk_part = Min_risk_part
            self.Capital_reinjection_rate = Capital_reinjection_rate
            self.Initial_funds = Initial_funds
            self.floor_percent = floor_percent
            self.multiplier = multiplier
            self.tipp = self.tipp(risk_returns,safe_asset,Lock_in,Min_risk_part \
                                  ,Capital_reinjection_rate,Initial_funds \
                                  ,floor_percent,multiplier,Rebalancement_frequency)

    def initiate_funds(self,risk_returns_mtx,safe_asset_mtx,Initial_funds,Rebalancement_frequency,floor_percent) :

        Goal = len(risk_returns_mtx) / Rebalancement_frequency   #This initiate the goal of xxx days of investment to today.
        Actualizer = (1+safe_asset_mtx[0] * Rebalancement_frequency / 252)**Goal
        floor_cap = Initial_funds * floor_percent / Actualizer
        return floor_cap,Goal


    def Matrix_Preparation(self,risk_return_mtx,safe_asset_mtx,floor_cap,Initial_funds):
        mtx= np.ones( len(risk_return_mtx) )
        floor_matrix = mtx * floor_cap
        Reference_cap_matrix , Fund_matrix  = mtx * Initial_funds , mtx *  Initial_funds
        Capital_reijection_matrix  =np.zeros( len(risk_return_mtx) )
        return floor_matrix , Reference_cap_matrix , Capital_reijection_matrix \
               , Fund_matrix


    def tipp(self,risk_returns,safe_asset,Lock_in,Min_risk_part \
             ,Capital_reinjection_rate,Initial_funds ,floor_percent,multiplier\
             ,Rebalancement_frequency) :

        risk_return_mtx = risk_returns.values.reshape(-1)
        safe_asset_mtx = safe_asset.values.reshape(-1)
        floor_cap,Goal = self.initiate_funds(risk_return_mtx,safe_asset_mtx,Initial_funds,Rebalancement_frequency,floor_percent)
        floor_matrix , Reference_cap_matrix , Capital_reijection_matrix , Fund_matrix = self.Matrix_Preparation(risk_return_mtx,safe_asset_mtx,floor_cap,Initial_funds)

        Fund_matrix , floor_matrix , Capital_reijection_matrix , Reference_cap_matrix,return_mtx = self.TPPI_calculator_MC(risk_return_mtx,floor_matrix , Reference_cap_matrix \
                                                                                                                ,Capital_reijection_matrix,Fund_matrix ,Lock_in \
                                                                                                                ,floor_percent,Goal,multiplier,Min_risk_part,safe_asset_mtx\
                                                                                                                ,Rebalancement_frequency,Capital_reinjection_rate)

        return Fund_matrix , floor_matrix , Capital_reijection_matrix ,\
                Reference_cap_matrix,risk_return_mtx[1:],safe_asset_mtx[1:], return_mtx



    @staticmethod
    @njit
    def TPPI_calculator_MC(risk_return_mtx,floor_matrix , Reference_cap_matrix , Capital_reijection_matrix \
                          ,Fund_matrix ,Lock_in,floor_percent,Goal,multiplier,Min_risk_part,safe_asset\
                          ,Rebalancement_frequency,Capital_reinjection_rate) :

        for i in range(1,floor_matrix.size) :

            if Fund_matrix[i-1] >= (1+Lock_in)*Reference_cap_matrix[i - 1]:
                Reference_cap_matrix[i] = Fund_matrix[ i - 1]
            else :
                Reference_cap_matrix[i] = Reference_cap_matrix[i - 1]
                
            floor_cap_update  = Fund_matrix[i-1] * floor_percent/((1+safe_asset[i-1] * Rebalancement_frequency)**Goal)

            if floor_cap_update > floor_matrix[i-1] : 
                floor_matrix[i] = floor_cap_update 
            else : 
                floor_matrix[i] = floor_matrix[i-1]
                    
            if Fund_matrix[i-1] < Reference_cap_matrix[i-1] * Capital_reinjection_rate :
                diff = Reference_cap_matrix[i-1] * Capital_reinjection_rate - Fund_matrix[i-1]
                Reference_cap_matrix[i] = Reference_cap_matrix[i-1] - diff
                Fund_matrix[i - 1]  = Fund_matrix[i - 1] + diff 
                Capital_reijection_matrix[i] = diff
            else : 
                Capital_reijection_matrix[i] = 0

            C = Fund_matrix[i-1] - floor_matrix[i-1]
            risk_asset = max(min(multiplier * C, Fund_matrix[i-1]),Min_risk_part * Fund_matrix[i-1]) 
            riskless_asset = Fund_matrix[i-1] - risk_asset
            Fund_matrix[i] = risk_asset * (1 + risk_return_mtx[i]) + riskless_asset * ( 1 + safe_asset[i])
            Goal = Goal - 1/Rebalancement_frequency
        
        return_mtx = (Fund_matrix[1:] - Fund_matrix[:-1])/Fund_matrix[:-1]

        return  Fund_matrix , floor_matrix , Capital_reijection_matrix , Reference_cap_matrix,return_mtx


import numpy as np 
import warnings
from pyinsurance.Metric_Generator.ratios import Sharpe_rat , Sortino_rat , Omega_rat, Modigliani_rat,Information_rat
from pyinsurance.Metric_Generator.returns_metrics import cagr , Cumulative_ret
from pyinsurance.Metric_Generator.Probabilistic_Sharpe_Ratio import probabilistic_sharpe_ratio
from pyinsurance.Metric_Generator.Standard_dev import stdeviation
from pyinsurance.Value_at_risk.Value_at_risk import VaR
from pyinsurance.ts.tipp_constructor import TIPP
from pyinsurance.ts.Drawdown import Drawdown_function
import pandas as pd 
from pyinsurance.Regressions.OLS_Basic import jensen_alpha_beta
warnings.filterwarnings('ignore')

class tipp_model:
    
    def __init__(self,risk_returns,safe_asset,lock_in\
                ,Min_risk_part,Capital_reinjection_rate,Initial_funds\
                ,floor_percent,multiplier , Benchmark_returns\
                ,Rebalancement_frequency = 252,theta = 0.01,Horizon = 1) -> None : 
        
        self.Runner(risk_returns,safe_asset,lock_in,\
                 Min_risk_part,Capital_reinjection_rate,Initial_funds\
                ,floor_percent,multiplier,Rebalancement_frequency)
        
        self.Benchmark_returns = Benchmark_returns.values.reshape(-1)[1:]
            
        
        self.Sharpe_ratio = Sharpe_rat(self.return_mtx , self.safe_asset_mtx ,\
                                            Rebalancement_frequency)
        
        self.Sortino_ratio = Sortino_rat(self.return_mtx , self.safe_asset_mtx ,\
                                              Rebalancement_frequency)
        
        self.Breadth = Information_rat(self.return_mtx, self.Benchmark_returns,\
                                             Rebalancement_frequency)
        
        self.PSR = probabilistic_sharpe_ratio(self.return_mtx, self.Benchmark_returns,\
                                              self.safe_asset_mtx,Rebalancement_frequency,\
                                              self.Sharpe_ratio)
        
        self.standard_deviation = stdeviation(self.return_mtx,Rebalancement_frequency)
        
        self.Cumulative_returns = Cumulative_ret(self.return_mtx)
        
        self.Omega_ratio = Omega_rat(self.return_mtx)
        
        self.CAGR = cagr(self.return_mtx,Rebalancement_frequency)
        
        self.Drawdown = Drawdown_function(self.return_mtx)
        
        self.Value_at_risk , self.Conditional_Value_at_risk = VaR(self.return_mtx, theta , Horizon)
        
        self.Modigliani_ratio = Modigliani_rat(self.return_mtx, self.Benchmark_returns,\
                                                self.safe_asset_mtx, Rebalancement_frequency\
                                                ,self.Sharpe_ratio)
            
        self.Beta , self.Alpha = jensen_alpha_beta(self.return_mtx , self.Benchmark_returns\
                                                        , Rebalancement_frequency )
        
    def Runner(self,risk_returns : np.ndarray ,safe_asset : np.ndarray ,lock_in : np.float64,\
                 Min_risk_part : np.float64 ,Capital_reinjection_rate : np.float64 ,Initial_funds : np.float64\
                ,floor_percent : np.float64,multiplier : np.float64 ,Rebalancement_frequency : np.float64):

        """
        Runs the TIPP class

        Parameters
        ----------
        risk_returns : np.ndarray
        safe_asset : np.ndarray
        lock_in : np.float64
        Min_risk_part : np.float64
        Capital_reinjection_rate : np.float64
        Initial_funds : np.float64
        floor_percent : np.float64
        multiplier : np.float64
        Rebalancement_frequency : np.float64

        Returns
        ----------

        Fund : np.ndarray , containing the cumulative fund Matrix
        floor : np.ndarray , containing the evolution of the floor level
        risk_returns_mtx : np.ndarray , containing the return of the original investment
        safe_asset_mtx : np.ndarray , containing the interest rates 
        return_mtx : np.ndarray , containing the return of the investment using a TIPP portfolio  
        """
        
        self.Fund , self.floor ,self.capital_reinjection,self.Reference_capital,\
        self.risk_returns_mtx , self.safe_asset_mtx , self.return_mtx  = TIPP(risk_returns,safe_asset,lock_in,\
                 Min_risk_part,Capital_reinjection_rate,Initial_funds\
                ,floor_percent,multiplier,Rebalancement_frequency).tipp
        pass 





import scipy
from scipy.stats import skew , kurtosis , norm 
import numpy as np 
from pyinsurance.Metric_Generator.ratios import Sharpe_rat

def estimated_sharpe_ratio_stdev(risk_returns,safe_asset,Rebalancement_frequency,Sharpe_Ratio):

        """
        Compute the standard deviation of the sharpe ratio across time

        Parameters
        ----------
        risk_returns : np.ndarray
        safe_asset : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, Sharpe ratio Standard deviation

        """

        N = risk_returns.size
        sk = skew(risk_returns)
        kurt = kurtosis(risk_returns)
        Sharpe_std = np.sqrt((1 + (0.5 * Sharpe_Ratio ** 2) - (sk * Sharpe_Ratio)\
                              + (((kurt - 3) / 4) * Sharpe_Ratio ** 2)) / (N - 1)) * Rebalancement_frequency
        return Sharpe_std
    
def probabilistic_sharpe_ratio(risk_returns, benchmark_returns,safe_asset,Rebalancement_frequency,Sharpe_Ratio):

        """
        Compute the Probabilistic Sharpe ratio 

        Parameters
        ----------
        risk_returns : np.ndarray
        benchmark_returns : np.ndarray
        safe_asset : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, PSR

        """

        Sharpe_std = estimated_sharpe_ratio_stdev(risk_returns,safe_asset,Rebalancement_frequency,Sharpe_Ratio)
        sharpe_ratio_benchmark = Sharpe_rat(benchmark_returns,safe_asset,Rebalancement_frequency)
        probabilistic_SR = norm.cdf((Sharpe_Ratio - sharpe_ratio_benchmark) / Sharpe_std)
        return probabilistic_SR


from pandas import DataFrame

from pyinsurance.data.utility import load_file


def load() -> DataFrame:

    """
    Load the sp500 data used in the examples
    Returns
    -------
    data : DataFrame
        Data set containing OHLC, adjusted close and the trading volume.
    """ 
    return load_file(__file__, "SPX.csv.gz")


import statsmodels.api as sm

def jensen_alpha_beta(risk_returns ,benchmark_returns,Rebalancement_frequency):

        """
        Compute the Beta and alpha of the investment under the CAPM 

        Parameters
        ----------
        risk_returns : np.ndarray
        benchmark_returns : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        np.float64,Beta,np.float64,Alpha  

        """


        benchmark_returns = sm.add_constant(benchmark_returns)
        model = sm.OLS(risk_returns,benchmark_returns).fit()
        alpha,beta = model.params[0] * Rebalancement_frequency , model.params[1]
        return beta,alpha


import numpy as np 
from pyinsurance.Metric_Generator.returns_metrics import Cumulative_ret

def Drawdown_function(risk_returns):

        """
        Compute the Drawdown serie 

        Parameters
        ----------
        risk_returns : np.ndarray

        Returns
        ----------

        np.ndarray,Drawdown 

        """

        C = Cumulative_ret(risk_returns)
        Drawdown = ( C / np.maximum.accumulate(C,axis = 0) - 1 ) * 100
        return Drawdown      


import numpy as np 
from arch import arch_model
from scipy.stats import t
 
def VaR(return_matrix, theta,Horizon): #500 datas needed 

        """
        Compute the Value-at-Risk and Conditional Value-at-Risk

        Parameters
        ----------
        risk_returns : np.ndarray
        theta : np.float64
        Horizon : np.int16

        Returns
        ----------

        np.ndarray,np.ndarray   VaR , CVaR

        """
        
        am     = arch_model(return_matrix * 100 , vol='Garch'\
                         , p=1 , o=1 , q=1,mean='AR', lags=1,dist='StudentsT')
        res    = am.fit(update_freq=1,disp='off')
        forecasts = res.forecast(horizon=Horizon)
        z      = (return_matrix - forecasts.mean.iloc[-1,0])/res.conditional_volatility
        z      = z[~np.isnan(z)] #Remove NaNs
        z_up   = - np.sort(z)
        params = t.fit(z_up)
        alpha  = params[0]
        q_99_t = np.sqrt((alpha-2)/alpha)*t.ppf(1-theta,alpha)
        VaR    = - (forecasts.mean.iloc[-1,0] + \
                 np.sqrt((forecasts.variance.iloc[-1,0]))*q_99_t)
        CVaR     = - (t.pdf(t.ppf(1-theta,alpha),alpha)/theta) * ( alpha+t.ppf(1-theta,alpha)**2)\
                /(alpha-1) * np.sqrt(forecasts.variance.iloc[-1,0]) - forecasts.mean.iloc[-1,0]
                
        return VaR,CVaR


import numpy as np 

def cagr(risk_returns,Rebalancement_frequency):

        """
        Compute the Compound Annual Growth Rate (CAGR)

        Parameters
        ----------
        risk_returns : np.ndarray
        Rebalancement_frequency : np.float64

        Returns
        ----------

        float64, CAGR

        """

        Value_earned = Cumulative_ret(risk_returns)[-1]
        N = risk_returns.size / Rebalancement_frequency
        cagr = Value_earned ** (1/N) - 1 
        return cagr
    
def Cumulative_ret(risk_returns):

        """
        Compute the Cumulative return across time 

        Parameters
        ----------
        risk_returns : np.ndarray

        Returns
        ----------

        np.ndarray, CAGR

        """

        return np.cumprod(1 + risk_returns)